from typing import Dict

from osmrx.apis_handler.core import ApiCore


class ErrorOverpassApi(ValueError):
    pass


class OverpassApi(ApiCore):

    __OVERPASS_URL: str = "https://www.overpass-api.de/api/interpreter"
    __OVERPASS_QUERY_PREFIX: str = "[out:json];"

    def _build_parameters(self, query: str) -> Dict:
        return {
            "data": f"{self.__OVERPASS_QUERY_PREFIX}{query}"
        }

    def query(self, query: str) -> Dict:
        parameters = self._build_parameters(query)
        return self.request_query(self.__OVERPASS_URL, parameters)



import os
import sys

import datetime

import logging


class Logger:

    _log_dir: str = "logs"
    _formatter = logging.Formatter(
        "%(asctime)s - %(name)-15s - %(levelname)-8s : %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    _log_date_file_format: str = datetime.datetime.now().strftime("%Y_%m_%d_%H%M%S")

    def __init__(
        self,
        parent=None,
        logger_name=None,
        logger_level="info",
        logger_dir=None,
        raise_error=False,
    ) -> None:
        if not parent:
            self.__logger_name = logger_name if logger_name else self.__class__.__name__

            self.logger = self.__create_logger(
                logger_level,
                f"/{logger_dir}/{self.__class__.__name__}"
                if logger_dir is not None
                else None,
            )
            self.raise_error = raise_error

    def __create_logger(self, logger_level, logger_dir):
        """
        create a logger

        :param logger_level: str
        :return:
        """

        levels = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "error": logging.ERROR,
            "critical": logging.CRITICAL,
        }

        logger_init = logging.getLogger(self.__logger_name)
        logger_init.setLevel(
            levels[logger_level] if logger_level in levels else logging.DEBUG
        )

        if not logger_init.handlers:
            logger_init.setLevel(
                levels[logger_level] if logger_level in levels else logging.DEBUG
            )

            handler = logging.StreamHandler(sys.stdout)
            handler.setLevel(
                levels[logger_level] if logger_level in levels else logging.DEBUG
            )
            handler.setFormatter(self._formatter)
            logger_init.addHandler(handler)

            if logger_dir is not None:
                log_path = f"{self._log_dir}_{self._log_date_file_format}"
                complete_log_path = f"{log_path}{os.path.dirname(logger_dir)}"

                if not os.path.isdir(complete_log_path):
                    os.makedirs(complete_log_path)

                handler_file = logging.FileHandler(f"{log_path}{logger_dir}.txt")
                handler_file.setLevel(
                    levels[logger_level] if logger_level in levels else logging.DEBUG
                )
                handler_file.setFormatter(self._formatter)
                logger_init.addHandler(handler_file)

        return logger_init



from typing import Dict
from typing import List

from shapely import Point
from shapely import LineString

from osmrx.globals.queries import OsmFeatureTypes

TOPO_FIELD: str = "topo_uuid"
ID_OSM_FIELD: str = "id"


class OverpassDataBuilder:
    __GEOMETRY_FIELD: str = "geometry"
    __LAT_FIELD: str = "lat"
    __LNG_FIELD: str = "lon"
    __FEATURE_TYPE_OSM_FIELD: str = "type"
    __PROPERTIES_OSM_FIELD: str = "tags"
    __OSM_URL_FIELD: str = "osm_url"
    __OSM_URL = "https://www.openstreetmap.org"

    _line_features = None

    def __init__(self, overpass_data: List[Dict]) -> None:

        self._prepare_data(overpass_data)

    def _prepare_data(self, raw_data: List[Dict]):
        self._grouped_features = {
            "points": filter(
                lambda x: x[self.__FEATURE_TYPE_OSM_FIELD] == OsmFeatureTypes.node.value,
                raw_data,
            ),
            "lines": filter(
                lambda x: x[self.__FEATURE_TYPE_OSM_FIELD] == OsmFeatureTypes.way.value,
                raw_data,
            )
        }

    def point_features(self) -> List[Dict]:
        point_features = []
        data = self._grouped_features["points"]
        for uuid_enum, feature in enumerate(data, start=1):
            geometry = Point(feature[self.__LNG_FIELD], feature[self.__LAT_FIELD])
            point_features.append(self._build_properties(uuid_enum, geometry, feature))
        return point_features

    def line_features(self) -> List[Dict]:
        line_features = []
        data = self._grouped_features["lines"]
        for uuid_enum, feature in enumerate(data, start=1):
            geometry = LineString(
                [(coordinates[self.__LNG_FIELD], coordinates[self.__LAT_FIELD])
                 for coordinates in feature[self.__GEOMETRY_FIELD]]
            )
            line_features.append(self._build_properties(uuid_enum, geometry, feature))
        return line_features

    def _build_properties(self, uuid_enum: int, geometry: Point | LineString, properties: Dict) -> Dict:
        tags_attributes = properties.get(self.__PROPERTIES_OSM_FIELD, {})
        return {
            **tags_attributes,
            ID_OSM_FIELD: str(properties[ID_OSM_FIELD]),
            self.__OSM_URL_FIELD: f"{self.__OSM_URL}/{properties[self.__FEATURE_TYPE_OSM_FIELD]}/"
                                  f"{properties[ID_OSM_FIELD]}",
            TOPO_FIELD: uuid_enum,  # do not cast to str, because topology processing need an int
            self.__GEOMETRY_FIELD: geometry
        }



from osmrx.apis_handler.models import Bbox, Location
from osmrx.apis_handler.overpass import OverpassApi
from osmrx.apis_handler.query_builder import QueryBuilder
from osmrx.graph_manager.graph_manager import GraphManager
from osmrx.helpers.logger import Logger
from osmrx.data_processing.overpass_data_builder import OverpassDataBuilder
from osmrx.globals.queries import OsmFeatureModes


class OsmNetworkCore(Logger):

    def __init__(self, osm_feature_mode: str):
        self._geo_filter = None
        self._query = None
        self._raw_data = None
        self._graph_manager: GraphManager | None = None

        super().__init__()

        self.osm_feature_mode = OsmFeatureModes[osm_feature_mode]

    @property
    def osm_feature_mode(self) -> OsmFeatureModes:
        """Return the osm feature"""
        return self._osm_feature_mode

    @osm_feature_mode.setter
    def osm_feature_mode(self, feature_mode: OsmFeatureModes) -> None:
        """Set the osm feature type to use"""
        self._osm_feature_mode = feature_mode
        self.logger.info(f"Building {self._osm_feature_mode} Data")
        self._build_query()
        if feature_mode != OsmFeatureModes.poi:
            self._graph_manager = GraphManager(self.logger, feature_mode)

    @property
    def geo_filter(self) -> Bbox | Location:
        """Return the geo filter"""
        return self._geo_filter

    @geo_filter.setter
    def geo_filter(self, geo_filter: Bbox | Location) -> None:
        """Set the geo filter"""
        self._geo_filter = geo_filter
        self.logger.info(f"From {self.geo_filter.location_name}")

    @property
    def query(self) -> str:
        """Return the query"""
        return self._query

    def _build_query(self) -> QueryBuilder:
        """Method must be implemented on children.
        Initialize the query. The geo filter must be set on the output"""
        self.logger.info("Building the query")
        return QueryBuilder(self.osm_feature_mode)

    def _execute_query(self) -> OverpassDataBuilder:
        """Execute the query with the Overpass API"""
        if self._query is not None:
            self.logger.info("Execute the query")
            raw_data = OverpassApi(logger=self.logger).query(self._query)
            return OverpassDataBuilder(raw_data["elements"])

    @property
    def data(self) -> None:
        raise NotImplemented

    def _execute(self):
        base_query = self._build_query()
        self._query = base_query.from_geo_filter(self.geo_filter)
        self._execute_query()


from typing import List, Dict

import rustworkx as rx
from rustworkx import PathLengthMapping
from shapely import MultiPoint, Point, concave_hull


class IsochronesFeature:
    _intervals = None
    _intervals_data = None

    def __init__(self, from_node: Point, precision: int = 1):
        self._from_node = from_node  # to be sure to have an ischrone on the from node
        self._precision = precision
        self._data = []

    def from_distances(self, intervals: List[int]):
        self.intervals = intervals
        self._intervals_data = {interval: list() for interval in self.intervals[::-1]}

    def from_times(self):
        # TODO support time
        ...

    def build(self, graph: rx.PyGraph | rx.PyDiGraph, shortest_path_lengths: PathLengthMapping):
        if self._intervals is None:
            raise ValueError("None interval defined")

        for indice, length in shortest_path_lengths.items():
            for interval in self.intervals[::-1]:
                if length < interval[-1]:
                    self._intervals_data[interval].append(
                        graph.get_node_data(indice))

        for interval, geom in self._intervals_data.items():
            if 0 in interval:
                geom.append(self._from_node)
            self._intervals_data[interval] = concave_hull(MultiPoint(geom), self._precision)
        self._clean_iso()

    def _clean_iso(self):
        isochrones_to_clean = list(self._intervals_data.items())
        for pos, isochrone in enumerate(isochrones_to_clean):
            interval = " to ".join(map(lambda x: str(x), isochrone[0]))
            geom = isochrone[-1]

            if pos < len(isochrones_to_clean) - 1:
                next_isochrone = isochrones_to_clean[pos + 1]
                self._data.append({"geometry": geom.difference(next_isochrone[-1]),
                                   "distance": interval})
            else:
                self._data.append({"geometry": geom,
                                   "distance": interval})

    @property
    def data(self) -> List[Dict]:
        return self._data

    @property
    def intervals(self):
        return self._intervals

    @intervals.setter
    def intervals(self, intervals: List[float | int]):
        self._intervals = list(zip(intervals, intervals[1:]))



from typing import Dict
from typing import TYPE_CHECKING

from requests_futures import sessions

from osmrx.helpers.misc import retry

if TYPE_CHECKING:
    from logging import Logger


class ErrorRequest(Exception):
    pass


class ApiCore:

    __NB_WORKER: int = 1
    __WORKED_STATUS_CODE: int = 200

    def __init__(self, logger: "Logger"):
        self.logger = logger

    def check_request_response(self, response) -> None:
        python_class_name = self.__class__.__name__
        response_code = response.result().status_code
        response_reason = f"{response_code}:{response.result().reason}"
        response_result_message = (
            f"{python_class_name}: Query {response_reason} "
            f"in {round(response.result().elapsed.total_seconds(), 2)} sec."
        )
        self.logger.info(f"{response_result_message}")

        if response_code != self.__WORKED_STATUS_CODE:
            raise ErrorRequest(
                f"{response_result_message}"
            )

    @retry(ErrorRequest, tries=4, delay=3, backoff=2, logger=None)
    def request_query(self, url: str, parameters: Dict) -> Dict:

        session = sessions.FuturesSession(max_workers=self.__NB_WORKER)
        response = session.get(url, params=parameters)

        self.check_request_response(response)
        return response.result().json()



from typing import List, Dict

from shapely import LineString

from osmrx.graph_manager.arc_feature import ArcFeature


class PathFeature:
    _graph = None
    _nodes_indices = None
    _features = None

    def __init__(self, graph, nodes_indice: List[int]):
        self._graph = graph
        self._nodes_indices = nodes_indice
        self._features = self._build_features()

    @property
    def path(self) -> LineString:
        """Return the path as a LineString geometry"""
        return LineString([self._graph.get_node_data(indice)
                           for indice in self._nodes_indices])

    def features(self) -> List[Dict]:
        """Return each LineStrings composing the path with their attributes"""
        return [feature.to_dict(with_attr=True) for feature in self._build_features()]

    def _build_features(self) -> List[ArcFeature]:
        """Get all the ArcFeature composing the path found"""
        return [self._graph.get_all_edge_data(*indices)[0]
                for indices in list(zip(self._nodes_indices, self._nodes_indices[1:]))]



from typing import Tuple, List, Dict

from osmrx.apis_handler.models import Bbox, Location
from osmrx.globals.queries import OsmFeatureModes
from osmrx.main.core import OsmNetworkCore


class OsmNetworkPoi(OsmNetworkCore):

    def __init__(self):
        super().__init__(osm_feature_mode=OsmFeatureModes.poi.value)

    def _execute_query(self) -> None:
        """Execute the query with the Overpass API"""
        raw_data = super()._execute_query()
        if raw_data is not None:
            self._raw_data = raw_data.point_features()

    @property
    def data(self) -> List[Dict]:
        return self._raw_data


class Pois(OsmNetworkPoi):
    """To manage Points of interest"""

    def __init__(self):
        super().__init__()

    def from_bbox(self, bounds: Tuple[float, float, float, float]):
        """Find Points of interest from bbox"""
        self.geo_filter = Bbox(*bounds)
        self._execute()

    def from_location(self, location: str):
        """Find Points of interest from location"""
        self.geo_filter = Location(location, logger=self.logger)
        self._execute()



from typing import Tuple, List, Dict, Any, Generator

from shapely import Point, MultiPoint
import rustworkx as rx

from osmrx.apis_handler.models import Location, Bbox
from osmrx.graph_manager.isochrones_feature import IsochronesFeature
from osmrx.graph_manager.path_feature import PathFeature
from osmrx.helpers.misc import buffer_point
from osmrx.main.core import OsmNetworkCore
from osmrx.topology.checker import TopologyChecker


class OsmNetworkRoads(OsmNetworkCore):

    def __init__(self, osm_feature_mode: str, nodes_to_connect: List[Dict] | None = None) -> None:
        super().__init__(osm_feature_mode=osm_feature_mode)
        self._graph_manager.connected_nodes = nodes_to_connect

    def _execute_query(self) -> None:
        """Execute the query with the Overpass API"""
        raw_data = super()._execute_query()
        if raw_data is not None:
            self._raw_data = raw_data.line_features()

    @property
    def additional_nodes(self) -> List[Dict] | None:
        """return the nodes defined to connect on the network"""
        return self._graph_manager.connected_nodes

    def _build_graph(self) -> None:
        """Fix topology issues for LineString features and build graph"""
        if self._raw_data is not None:
            self._graph_manager.features = self._raw_data

    def topology_checker(self) -> TopologyChecker:
        """Return topology data"""
        topology_result = TopologyChecker(self._graph_manager.features)
        self.logger.info("Topology analysis built.")
        return topology_result

    @property
    def data(self) -> List[Dict] | None:
        """Return the data"""
        if self._graph_manager.features is not None:
            return [feature.to_dict(with_attr=True) for feature in self._graph_manager.features]

    def graph(self) -> rx.PyGraph | rx.PyDiGraph:
        return self._graph_manager.graph

    def _execute(self):
        """Continue the execution by building the graph"""
        super()._execute()
        self._build_graph()


class Roads(OsmNetworkRoads):
    """To manage roads"""

    def __init__(self, mode: str, nodes_to_connect: List[Dict] | None = None):
        super().__init__(osm_feature_mode=mode, nodes_to_connect=nodes_to_connect)

    def from_bbox(self, bounds: Tuple[float, float, float, float]):
        """Find roads from bbox"""
        self.geo_filter = Bbox(*bounds)
        self._execute()

    def from_location(self, location: str):
        """Find roads from location"""
        self.geo_filter = Location(location, logger=self.logger)
        self._execute()


class GraphAnalysis(Roads):
    def __init__(self, mode: str, nodes_to_connect: List[Point]):
        # must be ordered
        nodes_to_connect = [{"topo_uuid": 999999 + enum, "geometry": node}
                            for enum, node in enumerate(nodes_to_connect)]
        super().__init__(mode=mode, nodes_to_connect=nodes_to_connect)

    def get_shortest_path(self) -> Generator[PathFeature, Any, None]:
        """Compute a shortest path from a node to an other node"""
        assert len(self.additional_nodes) > 1, "You need 2 points at least to compute a path"
        nodes = [node["geometry"] for node in self.additional_nodes]

        for from_point, to_point in zip(nodes, nodes[1:]):
            area = MultiPoint([from_point, to_point]).buffer(from_point.distance(to_point) / 2).bounds
            self.from_bbox(tuple([area[1], area[0], area[3], area[2]]))
            paths = self._graph_manager.compute_shortest_path(from_point, to_point)
            for path in paths:
                yield path
            self.logger.info(f"Shortest path(s) built from {from_point.wkt} to {to_point.wkt}.")

    def isochrones_from_distance(self, intervals: List[int], precision: float = 1.0) -> IsochronesFeature:
        """Compute isochrones from a node based on distances"""
        assert len(self.additional_nodes) == 1, "You need 1 point to compute an isochrone"
        nodes = [node["geometry"] for node in self.additional_nodes]

        for node in nodes:
            area = buffer_point(node.y, node.x, max(intervals) + 100).bounds
            self.from_bbox(tuple([area[1], area[0], area[3], area[2]]))
            isochrones = self._graph_manager.compute_isochrone_from_distance(node, intervals, precision)
            self.logger.info(f"Isochrones {isochrones.intervals} built from {node.wkt}.")
            return isochrones



from typing import Dict, List
from typing import TYPE_CHECKING

import copy

from shapely import Point

if TYPE_CHECKING:
    from osmrx.graph_manager.arc_feature import ArcFeature


class TopologyChecker:
    _features = None
    _directed = None

    def __init__(self, features: "List[ArcFeature]", directed: bool = False) -> None:
        self._features = features
        self._directed = directed  # TODO: seems not useful

    @property
    def lines_unchanged(self) -> List[Dict]:
        """Linestring without any changes"""
        unchanged = list(filter(lambda feature: feature.topo_status == "unchanged", self._features))
        return [feature.to_dict() for feature in unchanged]

    @property
    def lines_added(self) -> List[Dict]:
        """Linestring added"""
        lines_added = filter(lambda feature: feature.topo_status == "added", self._features)
        return [feature.to_dict() for feature in lines_added]

    @property
    def nodes_added(self) -> List[Dict]:
        """Nodes added on the graph"""
        nodes_added = []
        for node in self.lines_added:
            node["geometry"] = Point(node["geometry"].coords[0])
            nodes_added.append(node)
        return nodes_added

    @property
    def lines_split(self) -> List[Dict]:
        """Linestring split"""
        split = list(filter(lambda feature: feature.topo_status == "split", self._features))
        return [feature.to_dict() for feature in split]

    @property
    def intersections_added(self) -> List[Dict]:
        """Intersections nodes added"""
        intersections_added = []
        split = self.lines_split
        for node in split:
            for coords in [node["geometry"].coords[0], node["geometry"].coords[-1]]:
                feature_copy = copy.deepcopy(node)
                feature_copy["geometry"] = Point(coords)
                intersections_added.append(feature_copy)

        return intersections_added



from typing import List
from typing import TYPE_CHECKING
from dataclasses import dataclass

from shapely import Point
from shapely import Polygon
from shapely.geometry import shape

from osmrx.apis_handler.nominatim import NominatimApi

if TYPE_CHECKING:
    from osmrx.helpers.logger import Logger


class Bbox:
    """To manage bbox item"""

    _min_x = None
    _min_y = None
    _max_x = None
    _max_y = None

    def __init__(self, min_x: float, min_y: float, max_x: float, max_y: float) -> None:
        self._min_x = min_x
        self._min_y = min_y
        self._max_x = max_x
        self._max_y = max_y

    @property
    def location_name(self) -> str:
        """Cast to a string"""
        return f"{self._min_x}, {self._min_y}, {self._max_x}, {self._max_y}"


@dataclass
class NominatimItem:
    place_id: int
    licence: str
    _osm_id: int
    bounds: Bbox
    position: Point
    display_name: str
    polygon: Polygon

    @property
    def osm_id(self):
        """Build the osm_id by adding a fixed number: this is it!"""
        return self._osm_id + 3600000000


class Location:
    """To manage location name and attributes from Nominatim"""
    _location_name = None
    _values = None
    _limit = None

    def __init__(self, location_name: str, logger: "Logger", limit: int = 1) -> None:
        self.logger = logger
        self.location_name = location_name
        self._limit = limit

    @property
    def location_name(self) -> str:
        """return the location_name initialized"""
        return self._location_name

    @location_name.setter
    def location_name(self, location_name: str) -> None:
        """location_name setter"""
        self._location_name = location_name
        self.values = location_name

    @property
    def values(self) -> List[NominatimItem]:
        return self._values

    @values.setter
    def values(self, location_name: str) -> None:
        """return the nominatim data found"""
        self._values = []
        data_found = NominatimApi(self.logger, q=location_name, limit=self._limit).items
        for item in data_found:
            self._values.append(
                NominatimItem(
                    place_id=item["place_id"],
                    licence=item["licence"],
                    _osm_id=item["osm_id"],
                    bounds=Bbox(*item["boundingbox"]),
                    position=Point(item["lon"], item["lat"]),
                    display_name=item["display_name"],
                    polygon=shape(item["geojson"]),
                )
            )



from enum import Enum


class OsmFeatureModes(Enum):
    vehicle = "vehicle"
    pedestrian = "pedestrian"
    poi = "poi"


class OsmFeatureTypes(Enum):
    way = "way"
    node = "node"


# overpass queries
osm_queries: dict = {
    OsmFeatureModes.vehicle: {
        "query": 'way["highway"~"^('
        "motorway|"
        "trunk|"
        "primary|"
        "secondary|"
        "tertiary|"
        "unclassified|"
        "residential|"
        "pedestrian|"
        "motorway_link|"
        "trunk_link|"
        "primary_link|"
        "secondary_link|"
        "tertiary_link|"
        "living_street|"
        "service|"
        "track|"
        "bus_guideway|"
        "escape|"
        "raceway|"
        "road|"
        "bridleway|"
        "corridor|"
        "path"
        ')$"]["area"!~"."]({geo_filter});',
        "directed_graph": True,
        "feature_type": OsmFeatureTypes.way
    },
    OsmFeatureModes.pedestrian: {
        "query": 'way["highway"~"^('
        "motorway|"
        "cycleway|"
        "primary|"
        "secondary|"
        "tertiary|"
        "unclassified|"
        "residential|"
        "pedestrian|"
        "motorway_link|"
        "primary_link|"
        "secondary_link|"
        "tertiary_link|"
        "living_street|"
        "service|"
        "track|"
        "bus_guideway|"
        "escape|"
        "road|"
        "footway|"
        "bridleway|"
        "steps|"
        "corridor|"
        "path"
        ')$"]["area"!~"."]({geo_filter});',
        "directed_graph": False,
        "feature_type": OsmFeatureTypes.way
    },
    OsmFeatureModes.poi: {
        "query": 'node[~"^(amenity)$"~"('
            "bar|biergarten|cafe|drinking_water|fast_food|ice_cream|food_court|pub|restaurant|college|driving_school"
            "|kindergarten|language_school|library|music_school|school|sport_school|toy_library|university|"
            "bicycle_parking|bicycle_repair_station|bicycle_rental|boat_rental|boat_sharing|"
            "bus_station|car_rental|car_sharing|car_wash|vehicle_inspection|charging_station|ferry_terminal|fuel|taxi|"
            "atm|bank|bureau_de_change|baby_hatch|clinic|doctors|dentist|hospital|nursing_home|pharmacy|social_facility"
            "|veterinary|arts_centre|brothel|casino|cinema|community_centre|gambling|nightclub|planetarium|"
            "public_bookcase|social_centre|stripclub|studio|bicycle_parking|bicycle_rental|swingerclub|theatre"
            "|animal_boarding|animal_shelter|conference_centre|courthouse|"
            "crematorium|dive_centre|embassy|fire_station|give_box|internet_cafe|monastery|photo_booth|place_of_worship"
            "|police|post_box|post_depot|post_office|prison|public_bath|ranger_station|recycling|refugee_site|	"
            "sanitary_dump_station|shelter|shower|telephone|toilets|townhall|vending_machine|waste_basket"
            "|waste_disposal|waste_transfer_station|watering_place|water_point"
            ')"]({geo_filter});'
            'node[~"^(shop)$"~"."]({geo_filter});',
        "feature_type": OsmFeatureTypes.node
    }
}



import copy
from typing import List, Dict
from typing import TYPE_CHECKING

import rustworkx as rx
from shapely import Point

from osmrx.graph_manager.isochrones_feature import IsochronesFeature
from osmrx.graph_manager.path_feature import PathFeature
from osmrx.topology.cleaner import TopologyCleaner

from osmrx.globals.queries import OsmFeatureModes

if TYPE_CHECKING:
    from osmrx.graph_manager.arc_feature import ArcFeature


class GraphCore:

    def __init__(self, directed: bool = False):
        self._graph = None
        self._nodes_mapping = {}
        self._edges_mapping = {}
        self._directed = directed

        if directed:
            self._graph = rx.PyDiGraph(multigraph=False)  # noqa
        else:
            self._graph = rx.PyGraph(multigraph=False)  # noqa

    @property
    def graph(self) -> rx.PyDiGraph | rx.PyGraph:
        """Return the graph"""
        return self._graph

    def _add_nodes(self, node_value: Point) -> int:
        """Add a node"""
        if node_value not in self._nodes_mapping:
            self._nodes_mapping[node_value] = self.graph.add_node(node_value)
        return self._nodes_mapping[node_value]

    def add_edge(self, from_node_value: Point, to_node_value: Point, attr: "ArcFeature") -> None:
        """add ege based on 2 nodes"""
        from_indice = self._add_nodes(from_node_value)
        to_indice = self._add_nodes(to_node_value)
        if attr.topo_uuid not in self._edges_mapping:
            self._edges_mapping[attr.topo_uuid] = self.graph.add_edge(from_indice, to_indice, attr)
        else:
            raise ValueError(f"{attr.topo_uuid} edge exists: it should not!")

    def get_node_indice(self, node_value: Point) -> int | None:
        """Return the node value from indice"""
        if node_value in self._nodes_mapping:
            return self._nodes_mapping[node_value]
        raise ValueError(f"{node_value} node not found!")

    def compute_shortest_path(self, from_node: Point, to_node: Point) -> List[PathFeature]:
        """Compute a shortest path from a node to an ohter node"""
        edges = rx.dijkstra_shortest_paths(
            self.graph,
            self.get_node_indice(from_node),
            self.get_node_indice(to_node),
            weight_fn=lambda edge: edge.length)

        return [
            PathFeature(self.graph, node_indices)
            for _, node_indices in edges.items()
        ]

    def compute_isochrone_from_distance(self, from_node: Point, intervals: List[int],
                                        precision: float | int = 1.0) -> IsochronesFeature:
        """Compute isochrone from a distance interval"""
        intervals.sort()
        assert intervals[0] == 0, "The intervals must start with 0"

        from_node_indice = self.get_node_indice(from_node)
        weight_attribute_func = lambda edge: edge.length  # noqa

        if self._directed:
            edges = rx.digraph_dijkstra_shortest_path_lengths(self.graph, from_node_indice, weight_attribute_func)
        else:
            edges = rx.dijkstra_shortest_path_lengths(self.graph, from_node_indice, weight_attribute_func)

        iso_session = IsochronesFeature(from_node, precision)
        iso_session.from_distances(intervals)
        iso_session.build(self.graph, edges)
        return iso_session


class GraphManager(GraphCore):

    def __init__(self, logger, mode: OsmFeatureModes):
        self._mode = mode

        super().__init__(directed=self.directed)

        self.logger = logger  # TODO: add a logger if not set

        self._features = None
        self._connected_nodes = None

    @property
    def directed(self) -> bool:
        """Return the graph mode"""
        if self._mode == OsmFeatureModes.vehicle:
            return True
        return False

    @property
    def connected_nodes(self) -> List[Dict] | None:
        """return the connected nodes added"""
        return self._connected_nodes

    @connected_nodes.setter
    def connected_nodes(self, connected_nodes: List[Dict]):
        """Set the nodes to connect on the network"""
        self._connected_nodes = connected_nodes

    @property
    def features(self) -> "List[ArcFeature]":
        """Return the graph features from the graph"""
        return self._features

    @features.setter
    def features(self, network_data: List[Dict] | None):
        # TODO: improve name
        arc_features = TopologyCleaner(self.logger, network_data, self._connected_nodes, None).run()

        _ = [self._build_graph(arc_feature)
             for arc_feature in arc_features]

        self._features = self.graph.edges()
        self.logger.info("Graph built")

    def _build_graph(self, arc_feature: "ArcFeature"):

        self.add_edge(
            arc_feature.from_point,
            arc_feature.to_point,
            arc_feature
        )
        if self._mode == OsmFeatureModes.vehicle:
            if arc_feature.is_junction_or_roundabout():
                # do nothing
                return

            if not arc_feature.is_oneway():
                arc_feature_backward = copy.deepcopy(arc_feature)
                arc_feature_backward.direction = "backward"
                self.add_edge(
                    arc_feature_backward.from_point,
                    arc_feature_backward.to_point,
                    arc_feature_backward
                )



from osmrx.main.pois import Pois

from osmrx.main.roads import Roads
from osmrx.main.roads import GraphAnalysis

from osmrx.topology.cleaner import TopologyCleaner

from osmrx.graph_manager import graph_manager



from typing import Dict, List, Literal

from pyproj import Geod
from shapely import LineString, Point


class ArcFeature:
    __slots__ = ("_topo_uuid", "_geometry", "_topo_status", "_attributes", "_direction")

    def __init__(self, geometry: LineString):
        self._topo_uuid = None
        self._geometry = None
        self._topo_status = None
        self._direction = "forward"
        self._attributes = {}
        self._geometry = geometry

    @property
    def topo_uuid(self) -> str:
        return f"{self._topo_uuid}_{self._direction}"

    @topo_uuid.setter
    def topo_uuid(self, topo_uuid: str):
        self._topo_uuid = topo_uuid

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, direction: Literal["forward", "backward"]):
        self._direction = direction
        self._refresh_geometry()

    @property
    def geometry(self) -> LineString:
        return self._geometry

    def _refresh_geometry(self):
        if self._direction == "backward":
            self._geometry = LineString(self.coordinates[::-1])

    @property
    def coordinates(self) -> List[float]:
        return list(self._geometry.coords)

    @property
    def from_point(self) -> Point:
        return Point(self.coordinates[0])

    @property
    def to_point(self) -> Point:
        return Point(self.coordinates[-1])

    @property
    def topo_status(self) -> str:
        """Return the topology status"""
        return self._topo_status

    @topo_status.setter
    def topo_status(self, topo_status: Literal["added", "split", "unchanged"]):
        """Set the topology status"""
        self._topo_status = topo_status

    @property
    def length(self) -> float:
        """Compute the length of a wg84 LineString in meters"""
        return Geod(ellps="WGS84").geometry_length(self.geometry)

    @property
    def attributes(self) -> Dict[str, any]:
        """Return arc attributes"""
        return self._attributes

    @attributes.setter
    def attributes(self, attributes: Dict):
        """Return arc attributes"""
        self._attributes = attributes

    def to_dict(self, with_attr: bool = False) -> Dict[str, any]:
        """Return all the attributes as a dict"""
        main_attrs = {
            "topo_uuid": self.topo_uuid,
            "topo_status": self.topo_status,
            "geometry": self.geometry,
            "direction": self.direction,
        }
        if with_attr:
            return main_attrs | self.attributes
        return main_attrs

    def is_junction_or_roundabout(self) -> bool:
        """Check if arc is a junction or a roundabout"""
        return self.attributes.get("junction", None) in ["roundabout", "jughandle"]

    def is_oneway(self) -> bool:
        """Check if direction arc"""
        return self.attributes.get("oneway", None) == "yes"



from typing import TYPE_CHECKING

from osmrx.globals.queries import osm_queries
from osmrx.apis_handler.models import Bbox
from osmrx.apis_handler.models import Location

if TYPE_CHECKING:
    from osmrx.globals.queries import OsmFeatureModes


class ErrorQueryBuilder(Exception):
    pass


class QueryBuilder:
    
    _output_format = "out geom;(._;>;)"
    _area_tag_query: str = "area.searchArea"

    def __init__(self, mode: "OsmFeatureModes") -> None:
        self._osm_query = None
        self._query = None

        self._osm_query = osm_queries[mode]["query"]

    def from_geo_filter(self, geo_filter: Bbox | Location):
        if isinstance(geo_filter, Bbox):
            return self._from_bbox(geo_filter)
        elif isinstance(geo_filter, Location):
            return self._from_location(geo_filter)

    def _from_bbox(self, bbox: Bbox) -> str:
        """build a query from a bbox"""
        query = self._osm_query.format(geo_filter=bbox.location_name)
        return self._build_query(f"({query})")
    
    def _from_location(self, location: Location) -> str:
        """build a query from a location"""
        query = self._osm_query.format(geo_filter=self._area_tag_query)
        query = f"area({location.values[0].osm_id})->.searchArea;({query})"
        return self._build_query(query)

    def _build_query(self, query_with_geofilter: str) -> str:
        """Finalize the query with the output format"""
        return f"{query_with_geofilter};{self._output_format};"



from typing import Dict, List
from typing import Set
from typing import TYPE_CHECKING

from osmrx.apis_handler.core import ApiCore

if TYPE_CHECKING:
    from logging import Logger


class ErrorNominatimApi(ValueError):
    pass


class NominatimApi(ApiCore):

    nominatim_url = "https://nominatim.openstreetmap.org/search/?"

    query_parameter = "q"
    other_query_parameter: Set[str] = {
        "street",
        "city",
        "county",
        "state",
        "country",
        "postalcode",
    }
    format_parameter: Dict = {"format": "json", "polygon": "1", "polygon_geojson": "1"}

    def __init__(self, logger: "Logger", **params) -> None:
        _values = None
        super().__init__(logger=logger)

        parameters: Dict = self.__check_parameters(params)
        self.items = self.request_query(self.nominatim_url, parameters)

    def __check_parameters(self, input_parameters: Dict) -> Dict:

        if self.query_parameter in input_parameters:
            # clean arguments set
            for param_key in self.other_query_parameter:
                try:
                    del input_parameters[param_key]
                except KeyError:
                    pass

        elif not any(
            [
                input_key in self.other_query_parameter
                for input_key in input_parameters.keys()
            ]
        ):
            raise ErrorNominatimApi(
                f"{', '.join(self.other_query_parameter)} not found!"
            )

        input_parameters.update(self.format_parameter)

        return input_parameters

    @property
    def items(self) -> List[Dict]:
        return self._values

    @items.setter
    def items(self, data_found: List[Dict]) -> None:
        self._values = data_found




import time

from functools import wraps

from pyproj import Geod
from shapely import Point, Polygon


def retry(exceptions_to_check, tries: int = 4, delay: int = 3, backoff: int = 2, logger=None):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry
    """

    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)

                except exceptions_to_check as e:
                    msg = f"{str(e)}, Retrying in {mdelay} seconds..."
                    if logger:
                        logger.warning(msg)

                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry


def buffer_point(lon: float, lat: float, buffer_dist: float | int) -> Polygon:
    """Create a buffer from a 4326 point"""
    geod = Geod(ellps='WGS84')
    lon1, lat1, _ = geod.fwd(lon, lat, 0, buffer_dist)
    lon2, lat2, _ = geod.fwd(lon, lat, 180, buffer_dist)
    lon_diff, lat_diff = abs(lon1 - lon2), abs(lat1 - lat2)
    return Point(lat, lon).buffer(max(lon_diff, lat_diff))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['osmrx',
 'osmrx.apis_handler',
 'osmrx.data_processing',
 'osmrx.globals',
 'osmrx.graph_manager',
 'osmrx.helpers',
 'osmrx.main',
 'osmrx.topology']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.7.1,<4.0.0',
 'more-itertools>=9.1.0,<10.0.0',
 'requests-futures>=1.0.0,<2.0.0',
 'rtree>=1.0.1,<2.0.0',
 'rustworkx[mpl]>=0.12.1,<0.13.0',
 'scipy>=1.10.1,<2.0.0',
 'setuptools>=67.6.1,<68.0.0',
 'shapely>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'osmrx',
    'version': '0.0.5',
    'description': '',
    'long_description': '# OsmRx\n\nA geographic Python library to extract Open Street Map roads (and POIs) from a location or a bounding box, in order to create a graph thanks to Rustworkx. OsmRx is able to clean a network based on Linestring geometries and connect Point geometries. The graph built is able to process graph-analysis (shortest-path, isochrones...)\n\nCapabilities:\n* load data from a location name or a bounding box (roads and pois)\n* graph creation (and topology processing and cleaning)\n* shortest path\n* isochrone builder\n\n[![CI](https://github.com/amauryval/osmrx/actions/workflows/main.yml/badge.svg?branch=master)](https://github.com/amauryval/osmrx/actions/workflows/main.yml)\n[![codecov](https://codecov.io/gh/amauryval/osmrx/branch/master/graph/badge.svg)](https://codecov.io/gh/amauryval/osmrx)\n\n[![PyPI version](https://badge.fury.io/py/osmrx.svg)](https://badge.fury.io/py/osmrx)\n\nCheck the demo [here](https://amauryval.github.io/gdf2bokeh/)\n\n\n## How to install it ?\n\n### with pip\n\n```bash\npip install osmrx\n```\n\n## How to use it ?\n\nCheck the jupyter notebook here : TODO\nCheck the wiki: TODO\n\n### Get POIs\n\nTODO\n\n```python\n\n```\n\n### Get Roads\n\nTODO\n\n```python\n\n```\n\n\n### Compute a shortest path\n\nTODO\n\n```python\n\n```\n\n\n### Compute an Isochrone\n\nTODO\n\n```python\n\n```\n',
    'author': 'amauryval',
    'author_email': 'amauryval@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.11.0',
}


setup(**setup_kwargs)



import itertools
from typing import Tuple, Generator, Any
from typing import List
from typing import Dict
from typing import Optional
from typing import Set
from typing import Union
from typing import Iterator

from numpy import ndarray
from scipy import spatial

from shapely.geometry import LineString

import rtree

import numpy as np

from collections import Counter

from more_itertools import split_at

import concurrent.futures

from osmrx.data_processing.overpass_data_builder import TOPO_FIELD, ID_OSM_FIELD
from osmrx.graph_manager.arc_feature import ArcFeature


class NetworkTopologyError(Exception):
    pass


class LineBuilder:
    __TOPOLOGY_TAG_SPLIT: str = "split"

    __CLEANING_FILED_STATUS: str = "topology"

    __LINESTRING_SEPARATOR: str = "_"

    def __init__(self, feature: Dict, intersection_nodes: set[tuple[float, float]],
                 interpolate_level: int | None = None):
        self._feature = feature
        del self._feature["geometry"]
        self._coordinates = self._feature.pop("coordinates")
        self._unique_coordinates = set(self._coordinates)
        self._intersection_nodes = intersection_nodes
        self._interpolate_level = interpolate_level

        self._output = []

    def build_features(self) -> List[ArcFeature]:
        if not self.is_line_valid():
            return []

        geometry_lines = self.split_line_at_intersections(
            self._coordinates, self.intersections_points()
        )
        if len(geometry_lines) > 1:
            self._feature[self.__CLEANING_FILED_STATUS] = self.__TOPOLOGY_TAG_SPLIT

            data = []
            for suffix_id, line_coordinates in enumerate(geometry_lines):
                feature_copy = self.feature_copy()
                feature_copy["topo_uuid"] = f"{feature_copy['topo_uuid']}_{suffix_id}"
                data.append([feature_copy, line_coordinates])

                # feature_updated[self.__COORDINATES_FIELD] = line_coordinates
                self._direction_processing(feature_copy, line_coordinates)
            # with concurrent.futures.ThreadPoolExecutor() as executor:
            #     executor.map(lambda f: self._direction_processing(*f), data)
        else:
            self._direction_processing(self._feature, geometry_lines[0])

        return self._output

    def _direction_processing(self, input_feature: Dict, new_coordinates: list[tuple[float, float]]):

        if self._interpolate_level:
            new_coords = list(
                self._split_line(new_coordinates, self._interpolate_level)
            )
            new_lines_coords = list(zip(new_coords, new_coords[1:]))

            for idx, sub_line_coords in enumerate(new_lines_coords):
                feature_copy = dict(input_feature)
                self.__proceed_direction_geom(feature_copy, sub_line_coords, idx)

        else:
            self.__proceed_direction_geom(input_feature, new_coordinates, None)

    def __proceed_direction_geom(self, input_feature, sub_line_coords, idx: int | None):
        # TODO maybe useless: check parent method

        position = ""
        if idx:
            position = f"_{idx}"

        new_feature = ArcFeature(LineString(sub_line_coords))
        new_feature.topo_uuid = f"{input_feature.pop('topo_uuid')}{position}"
        new_feature.topo_status = input_feature.pop(self.__CLEANING_FILED_STATUS)
        new_feature.attributes = input_feature

        self._output.append(new_feature)

    @staticmethod
    def _split_line(coordinates: list[tuple[float, float]], interpolation_level: int) -> ndarray:
        return interpolate_curve_based_on_original_points(
            np.array(coordinates), interpolation_level,
        )

    def feature_copy(self) -> Dict:
        return dict(self._feature)

    def intersections_points(self) -> Set[Tuple[float, float]]:
        """Return intersections points matching with the feature"""
        return self._unique_coordinates.intersection(self._intersection_nodes)

    def is_line_valid(self) -> True:
        # meaning that there is none point or line length is equals to 0
        return not len(self._unique_coordinates) <= 1

    def split_line_at_intersections(self, coordinates: List[Tuple[float, float]],
                                    points_intersections: Set[Tuple[float, float]]) -> List[List[Tuple[float, float]]]:

        if len(points_intersections) > 0:

            # split coordinates found at intersection to respect the topology
            first_value, *middle_coordinates_values, last_value = coordinates
            for point_intersection in points_intersections:

                if point_intersection in middle_coordinates_values:
                    # we get the middle values from coordinates to avoid to catch the first and last value when editing

                    # duplicate the intersection point
                    index: int = middle_coordinates_values.index(point_intersection)
                    middle_coordinates_values[index:index] = [point_intersection]
                    # add an _ to split the line at the intersection point index
                    middle_coordinates_values[index + 1:index + 1] = self.__LINESTRING_SEPARATOR

            coordinates = [first_value] + middle_coordinates_values + [last_value]
            coordinates_updated = list(split_at(coordinates, lambda x: x == self.__LINESTRING_SEPARATOR))
        else:
            coordinates_updated = list([coordinates])

        return coordinates_updated


class TopologyCleaner:

    # if increased, the node connections will be better, but will generate more feature
    __INTERPOLATION_LEVEL: int = 7
    __NB_OF_NEAREST_LINE_ELEMENTS_TO_FIND: int = 10

    __NUMBER_OF_NODES_INTERSECTIONS: int = 2

    __CLEANING_FILED_STATUS: str = "topology"
    __GEOMETRY_FIELD: str = "geometry"
    __COORDINATES_FIELD: str = "coordinates"

    __TOPOLOGY_TAG_ADDED: str = "added"
    __TOPOLOGY_TAG_UNCHANGED: str = "unchanged"

    def __init__(
        self,
        logger,  # TODO: add a logger if not set
        network_data: List[Dict],
        additional_nodes: Optional[List[Dict]],
        interpolation_line_level: int | None = None,  # 4 is a good value
    ) -> None:

        self.logger = logger
        self.logger.info("Network cleaning...")

        self._network_data: Union[List[Dict], Dict] = network_data
        self._interpolation_line_level = interpolation_line_level  # link to __INTERPOLATION_LINE_LEVEL

        self._additional_nodes = additional_nodes
        if self._additional_nodes is None:
            self._additional_nodes: Dict = {}

        self.__FIELD_ID = TOPO_FIELD  # have to be an integer.. thank rtree...
        self._original_field_id = ID_OSM_FIELD

        self._intersections_found: Optional[Set[Tuple[float, float]]] = None
        self.__connections_added: Dict = {}

    def run(self) -> Generator[ArcFeature, Any, None]:
        self._prepare_data()

        # connect all the added nodes
        if len(self._additional_nodes) > 0:
            self.compute_added_node_connections()

        # find all the existing intersection from coordinates
        intersections_found = self.find_intersections_from_ways()

        self.logger.info("Build lines")

        for feature in self._network_data.values():
            for feature_built in LineBuilder(feature, intersections_found,
                                             self._interpolation_line_level).build_features():
                yield feature_built

    def _prepare_data(self):

        self._network_data = {
            feature[self.__FIELD_ID]: {
                self.__COORDINATES_FIELD: feature[self.__GEOMETRY_FIELD].coords[:],
                self.__CLEANING_FILED_STATUS: self.__TOPOLOGY_TAG_UNCHANGED,
                **feature,
            }
            for feature in self._network_data
        }
        self._additional_nodes = {
            feature[self.__FIELD_ID]: {
                self.__COORDINATES_FIELD: feature[self.__GEOMETRY_FIELD].coords[0],
                **feature,
            }
            for feature in self._additional_nodes
        }

    def compute_added_node_connections(self):
        self.logger.info("Starting: Adding new nodes on the network")

        self.logger.info("Find nearest line for each node")
        node_keys_by_nearest_lines_filled = (
            self.__find_nearest_line_for_each_key_nodes()
        )

        self.logger.info("Split line")
        # for nearest_line_key in node_keys_by_nearest_lines_filled:
        #     self.split_line(nearest_line_key)
        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(self.split_line, node_keys_by_nearest_lines_filled)

        self._network_data: Dict = self._network_data | self.__connections_added

    def split_line(self, node_key_by_nearest_lines):
        nearest_line_content = self.__node_by_nearest_lines[node_key_by_nearest_lines]
        default_line_updater = self.proceed_nodes_on_network(
            (node_key_by_nearest_lines, nearest_line_content)
        )
        if default_line_updater is not None:
            self.insert_new_nodes_on_its_line(default_line_updater)

    def insert_new_nodes_on_its_line(self, item):
        original_line_key = item["original_line_key"]
        end_points_found = item["end_points_found"]

        linestring_with_new_nodes = self._network_data[original_line_key][self.__COORDINATES_FIELD]
        linestring_with_new_nodes.extend(end_points_found)
        linestring_with_new_nodes = set(linestring_with_new_nodes)

        # build new LineStrings
        linestring_linked_updated = list(filter(lambda x: x in linestring_with_new_nodes, item["interpolated_line"]))

        self._network_data[original_line_key][self.__COORDINATES_FIELD] = linestring_linked_updated

    def proceed_nodes_on_network(self, nearest_line_content):
        nearest_line_key, node_keys = nearest_line_content

        interpolated_line_coords = interpolate_curve_based_on_original_points(
            np.array(self._network_data[nearest_line_key][self.__COORDINATES_FIELD]),
            self.__INTERPOLATION_LEVEL,
        )
        line_tree = spatial.cKDTree(interpolated_line_coords)
        interpolated_line_coords_rebuilt = list(map(tuple, interpolated_line_coords))

        nodes_coords = [
            self._additional_nodes[node_key][self.__COORDINATES_FIELD]
            for node_key in node_keys
        ]
        _, nearest_line_object_idxes = line_tree.query(nodes_coords)
        end_points_found = [
            interpolated_line_coords_rebuilt[nearest_line_key]
            for nearest_line_key in nearest_line_object_idxes
        ]

        connections_coords = zip(node_keys, list(zip(nodes_coords, end_points_found)))
        connections_coords_valid = filter(lambda x: len(set(x[-1])) > 0, connections_coords)
        for node_key, connection in connections_coords_valid:

            # to split line at node (and also if node is on the network). it builds intersection used to split lines
            # additional are converted to lines
            self.__connections_added[f"from_node_id_{node_key}"] = {
                self.__COORDINATES_FIELD: connection,
                self.__GEOMETRY_FIELD: connection,
                self.__CLEANING_FILED_STATUS: self.__TOPOLOGY_TAG_ADDED,
                self.__FIELD_ID: f"{self.__TOPOLOGY_TAG_ADDED}_{node_key}",
                self._original_field_id: f"{self.__TOPOLOGY_TAG_ADDED}_{node_key}",
            }

        return {
            "interpolated_line": interpolated_line_coords_rebuilt,
            "original_line_key": nearest_line_key,
            "end_points_found": end_points_found,
        }

    def find_intersections_from_ways(self) -> Set[Tuple[float, float]]:
        self.logger.info("Starting: Find intersections")
        all_coord_points = Counter(
            itertools.chain.from_iterable([
                feature[self.__COORDINATES_FIELD]
                for feature in self._network_data.values()
            ]),
        )

        intersections_found = dict(
            filter(
                lambda x: x[1] >= self.__NUMBER_OF_NODES_INTERSECTIONS,
                all_coord_points.items(),
            )
        ).keys()
        self.logger.info("Done: Find intersections")

        return set(intersections_found)

    def __rtree_generator_func(
        self,
    ) -> Iterator[Tuple[int, Tuple[str, str, str, float], None]]:
        for fid, feature in self._network_data.items():
            # fid is an integer
            yield fid, feature[self.__GEOMETRY_FIELD].bounds, None

    def __find_nearest_line_for_each_key_nodes(self) -> Iterator[int]:
        # find the nearest network arc to interpolate
        self._tree_index = rtree.index.Index(self.__rtree_generator_func())

        # find the nearest line
        self.__node_by_nearest_lines = dict(
            (key, []) for key in self._network_data.keys()
        )

        # not working because rtree cannot be MultiThreaded
        # with concurrent.futures.ThreadPoolExecutor(4) as executor:
        #     executor.map(self.__get_nearest_line, self._additional_nodes.items())
        for node_info in self._additional_nodes.items():
            self.__get_nearest_line(node_info)

        node_keys_by_nearest_lines_filled = filter(
            lambda x: len(self.__node_by_nearest_lines[x]) > 0,
            self.__node_by_nearest_lines,
        )

        return node_keys_by_nearest_lines_filled

    def __get_nearest_line(self, node_info: Tuple[int, Dict]) -> None:
        node_uuid, node = node_info
        distances_computed: List[Tuple[float, int]] = []
        node_geom = node[self.__GEOMETRY_FIELD]

        for index_feature in self._tree_index.nearest(node_geom.bounds, self.__NB_OF_NEAREST_LINE_ELEMENTS_TO_FIND):
            line_geom = self._network_data[index_feature][self.__GEOMETRY_FIELD]
            distance_from_node_to_line = node_geom.distance(line_geom)
            distances_computed.append((distance_from_node_to_line, index_feature))
            if distance_from_node_to_line == 0:
                # means that the node is on the network, looping is not necessary anymore
                break

        _, line_min_index = min(distances_computed)
        self.__node_by_nearest_lines[line_min_index].append(node_uuid)


def interpolate_curve_based_on_original_points(values: np.array, interpolation_factor: int) -> np.ndarray:
    # Convert values to a 2D array (if necessary) and remove single-dimensional entries
    values = np.squeeze(np.atleast_2d(values))

    # Compute the total number of points after interpolation
    n_interpolated_points = len(values) * interpolation_factor

    # Create an array with indices of the original points
    original_indices = np.arange(0, n_interpolated_points, interpolation_factor)

    # Create an array with indices of the new points
    new_indices = np.arange(n_interpolated_points - 1)

    # Compute the interpolated x and y coordinates
    x_interpolated = np.interp(new_indices, original_indices, values[:, 0])
    y_interpolated = np.interp(new_indices, original_indices, values[:, 1])

    # Combine the x and y coordinates into a single array and return it
    interpolated_values = np.column_stack((x_interpolated, y_interpolated))
    return interpolated_values
